/*	$NetBSD$	*/

/*-
 * Copyright (c) 2009 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _STAR_INTR_H_
#define _STAR_INTR_H_

#define ARM_IRQ_HANDLER	_C_LABEL(star_intr_dispatch)

#ifndef _LOCORE

#include <arm/cpu.h>
#include <arm/armreg.h>
#include <arm/cpufunc.h>

#include <arm/star/starreg.h>

extern uint32_t star_intr_mask[];
extern volatile uint32_t star_intr_enabled;
extern volatile uint32_t star_intr_pending;

static inline void __attribute__((__unused__))
star_set_intrmask(void)
{
	STAR_REG_WRITE32(EQUULEUS_INT_CLEAR, star_intr_enabled);
	STAR_REG_WRITE32(EQUULEUS_INT_MASK, ~star_intr_enabled);
	STAR_REG_WRITE32(EQUULEUS_INT_MASKCLEAR, star_intr_enabled);
}

static inline void __attribute__((__unused__))
star_splx(int ipl)
{
	int oldirqstate, hwpend;

	/* Don't let the compiler re-order this code with preceding code */
	__insn_barrier();

	set_curcpl(ipl);

	hwpend = (star_intr_pending) & ~star_intr_mask[ipl];
	if (hwpend != 0) {
		oldirqstate = disable_interrupts(I32_bit);
		star_intr_enabled |= hwpend;
		star_set_intrmask();
		restore_interrupts(oldirqstate);
	}

#ifdef __HAVE_FAST_SOFTINTS
	cpu_dosoftints();
#endif
}

static inline int __attribute__((__unused__))
star_splraise(int ipl)
{
	int old;

	old = curcpl();
	set_curcpl(ipl);

	/* Don't let the compiler re-order this code with subsequent code */
	__insn_barrier();

	return old;
}

static inline int __attribute__((__unused__))
star_spllower(int ipl)
{
	int old;

	old = curcpl();
	star_splx(ipl);

	return old;
}

void splx(int);
int _spllower(int);
int _splraise(int);
void _setsoftintr(int);

#if !defined(EVBARM_SPL_NOINLINE)
#define splx(ipl)		star_splx(ipl)
#define _spllower(ipl)		star_spllower(ipl)
#define _splraise(ipl)		star_splraise(ipl)
#define _setsoftintr(ipl)	star_setsoftintr(ipl)
#endif /* EVBARM_SPL_NOINLINE */

void star_intr_init(void);
void star_intr_dispatch(void *);
void *star_intr_establish(int, int, int (*)(void *), void *);
void star_intr_disestablish(void *);

#endif /* _LOCORE */

#endif /* _STAR_INTR_H_ */
