/*	$NetBSD$	*/

/*-
 * Copyright (c) 2009 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <machine/bus.h>

#include <arm/star/starreg.h>
#include <arm/star/starvar.h>
#include <arm/star/if_gecreg.h>

struct star_devtable star_soc_devtable[] = {
	{
		.paddr = STR8100_FLASH_SRAM_MEMORY_BANK0,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_FLASH_SRAM_MEMORY_BANK1,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_FLASH_SRAM_MEMORY_BANK2,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_FLASH_SRAM_MEMORY_BANK3,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_IDE_DEVICE_REGISTER_SPACE,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_SPI_SERIAL_FLASH_MEMORY,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR9100_STATIC_MEMORY_CONTROL_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STR9100_SDRAM_CONTROL_REGISTER,
		.size = 0x00000030,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STRx100_GENERIC_DMA_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STR8100_NIC_REGISTER,
		.size = GEC_REG_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR9100_SWITCH_AND_HNAT_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STR8100_SPI_PCM_TWI_I2S_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_SDRAM_CONTROL_REGISTER,
		.size = 0x00000040,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_STATIC_MEMORY_CONTROL_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_IDE_CONTROL_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STRx100_MISC_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_POWER_MANAGEMENT_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_UART0_REGISTER,
		.size = STAR_UART_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STR8100_UART1_REGISTER,
		.size = STAR_UART_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STRx100_TIMER_REGISTER,
		.size = STRx100_TIMER_REGSIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_WDT_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_RTC_REGISTER,
		.size = STAR_RTC_REGSIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_GPIOA_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STR8100_GPIOB_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR9100_INTERRUPT_CONTROL_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STR9100_PCMCIA_CONTROL_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STR9100_PCMCIA_ATTRIBUTE_MEMORY,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STR9100_PCMCIA_COMMON_MEMORY,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STR9100_PCMCIA_IO_SPACE,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 0, .has_str9100 = 1
	},
	{
		.paddr = STRx100_PCI_CONFIGURATION_DATA_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_PCI_CONFIGURATION_ADDR_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_PCI_IO_SPACE,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_PCI_MEMORY_SPACE,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_USB11_CONFIGURATION_REGISTER,
		.size = STAR_USB11_CONF_REGSIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_USB11_OPERATION_REGISTER,
		.size = STAR_USB11_OPERATION_REGSIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_USB20_CONFIGURATION_REGISTER,
		.size = STAR_USB20_CONF_REGSIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STRx100_USB20_OPERATION_REGISTER,
		.size = STAR_USB20_OPERATION_REGSIZE,
		.has_str8100 = 1, .has_str9100 = 1
	},
	{
		.paddr = STR8100_USB11_20_DEVICE_REGISTER,
		.size = STAR_IOADDR_SIZE,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		.paddr = STR8100_INTERRUPT_CONTROL_REGISTER,
		.size = 0x00001000,
		.has_str8100 = 1, .has_str9100 = 0
	},
	{
		0
	}
};


int
star_get_memsize(int use_phys)
{
	int memsize;
	uint32_t val;

	memsize = 0;

	/*
	 * Use physical address, because this function would be called
	 * when virtual I/O address is not mapped yet.
	 */

#if defined(STAR_EQUULEUS) && defined(STAR_ORION)
	val = use_phys ?
	    STAR_REG_READ32(STAR_IOADDR_V2P(STAR_MISC_PCI_ID_CAPABILITY)) :
	    STAR_REG_READ32(STAR_MISC_PCI_ID_CAPABILITY);

	if (val == 0x8131eeee) {
#endif
#ifdef STAR_EQUULEUS
		/* STR8100 */
		if (use_phys) {
			val = STAR_REG_READ32(
			    STAR_IOADDR_V2P(EQUULEUS_SDRAMCTRL_MEM_CFG_REG));
		} else {
			val = STAR_REG_READ32(EQUULEUS_SDRAMCTRL_MEM_CFG_REG);
		}

		switch (val & EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_MASK) {
		case EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_128Mb:
			memsize = 128 * 1024 * 1024 / 8;
			break;
		case EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_256Mb:
			memsize = 256 * 1024 * 1024 / 8;
			break;
		case EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_512Mb:
			memsize = 512 * 1024 * 1024 / 8;
			break;
		case EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_1Gb:
			memsize = 1024 * 1024 * 1024 / 8;
			break;
		case EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_16Mb:
			memsize = 16 * 1024 * 1024 / 8;
			break;
		case EQUULEUS_SDRAMCTRL_MEM_CFG_MEMSZ_64Mb:
			memsize = 64 * 1024 * 1024 / 8;
			break;
		}
#endif
#if defined(STAR_EQUULEUS) && defined(STAR_ORION)
	} else {
#endif
#ifdef STAR_ORION
		/* STR9100 */
		if (use_phys) {
			val = STAR_REG_READ32(
			    STAR_IOADDR_V2P(ORION_SDRAMCTRL_BANK_CFG_REG));
		} else {
			val = STAR_REG_READ32(ORION_SDRAMCTRL_BANK_CFG_REG));
		}

		switch (val & ORION_SDRAMCTRL_BANK_CFG_BNK_DSZ_MASK) {
		case ORION_SDRAMCTRL_BANK_CFG_BNK_DSZ_64Mb:
			memsize = 64 * 1024 * 1024 / 8;
			break;
		case ORION_SDRAMCTRL_BANK_CFG_BNK_DSZ_128Mb:
			memsize = 128 * 1024 * 1024 / 8;
			break;
		case ORION_SDRAMCTRL_BANK_CFG_BNK_DSZ_256Mb:
			memsize = 256 * 1024 * 1024 / 8;
			break;
		case ORION_SDRAMCTRL_BANK_CFG_BNK_DSZ_512Mb:
			memsize = 512 * 1024 * 1024 / 8;
			break;
		}
#endif
#if defined(STAR_EQUULEUS) && defined(STAR_ORION)
	}
#endif

	return memsize;
}

