//#error THIS DRIVER IS WORKING STATE
/*	$NetBSD$	*/

/*-
 * Copyright (c) 2009 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/device.h>
#include <sys/errno.h>

#include <arm/star/starreg.h>
#include <arm/star/starvar.h>

#include <dev/i2c/i2cvar.h>

struct startwi_softc {
	device_t sc_dev;
	bus_addr_t sc_addr;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	void *sc_ih;

	struct i2c_controller sc_i2c;
	kmutex_t sc_buslock;
};

static int startwi_match(device_t, struct cfdata *, void *);
static void startwi_attach(device_t, device_t, void *);
static void startwi_init(struct startwi_softc *);
static int startwi_intr(void *);

CFATTACH_DECL_NEW(startwi, sizeof(struct startwi_softc),
    startwi_match, startwi_attach, NULL, NULL);


/* ARGSUSED */
static int
startwi_match(device_t parent, struct cfdata *match, void *aux)
{
	struct star_attach_args *sa;

	sa = aux;
	sa->sa_size = EQUULEUS_TWI_REGSIZE;

	return 1;
}

/* ARGSUSED */
static void
startwi_attach(device_t parent, device_t self, void *aux)
{
	struct startwi_softc *sc;
	struct star_attach_args *sa;
	struct i2cbus_attach_args iba;

	sa = aux;
	sc = device_private(self);
	sc->sc_dev = self;
	sc->sc_iot = sa->sa_iot;
	sc->sc_addr = sa->sa_addr;

	aprint_normal(": Two-Wire Serial Interface\n");
	aprint_naive("\n");

	if (bus_space_map(sc->sc_iot, sc->sc_addr, sa->sa_size, 0,
	    &sc->sc_ioh)) {
		aprint_error(": can't map registers\n");
		return;
	}

	startwi_init(sc);

	mutex_init(&sc->sc_buslock, MUTEX_DEFAULT, IPL_NONE);
	sc->sc_ih = star_intr_establish(STAR_IRQ_TWI, IPL_SERIAL,
	    startwi_intr, sc);


	sc->sc_i2c.ic_cookie = sc;
	sc->sc_i2c.ic_acquire_bus = NULL;
	sc->sc_i2c.ic_release_bus = NULL;
	sc->sc_i2c.ic_send_start = NULL;
	sc->sc_i2c.ic_send_stop = NULL;
	sc->sc_i2c.ic_initiate_xfer = NULL;
	sc->sc_i2c.ic_read_byte = NULL;
	sc->sc_i2c.ic_write_byte = NULL;
	sc->sc_i2c.ic_exec = NULL;

	iba.iba_tag = &sc->sc_i2c;
	config_found_ia(sc->sc_dev, "i2cbus", &iba, iicbus_print);
}

static void
startwi_init(struct startwi_softc *sc)
{
	uint32_t val;

	/* SPI/PCM/I2S/TWI Controler Enable */
	val = STAR_REG_READ32(STAR_CLKPWR_CLKGATE1_REG);
	STAR_REG_WRITE32(STAR_CLKPWR_CLKGATE1_REG,
	    val | STAR_CLKPWR_CLKGATE1_PCLK_P2S);

	/* activate SPI/PCM/I2S/TWI */
	val = STAR_REG_READ32(STAR_CLKPWR_SOFTRST_REG);
	STAR_REG_WRITE32(STAR_CLKPWR_SOFTRST_REG,
	    val | STAR_CLKPWR_SOFTRST_P2S);
}

static int
startwi_intr(void *arg)
{
	struct startwi_softc *sc;

	sc = (struct startwi_softc *)arg;

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_INT_STATUS_REG,
	    0xffffffff);	/* all clear */

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_INT_ENABLE_REG,
	    EQUULEUS_TWI_INT_ACTDONE |
	    EQUULEUS_TWI_INT_BUSERR);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_CTRL_REG,
	    EQUULEUS_TWI_CTRL_EN);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_TIMEOUT_REG,
	    EQUULEUS_TWI_TIMEOUT_CLKDIV(0x3e) |
	    EQUULEUS_TWI_TIMEOUT_OUT_EN |
	    EQUULEUS_TWI_TIMEOUT_OUT_VA(0x40));

//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_SLAVE_ADDR_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_WR_DATA_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, EQUULEUS_TWI_RD_DATA_REG, 0);

	return 1;
}
