/*	$NetBSD$	*/

/*-
 * Copyright (c) 2009 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/device.h>
#include <sys/errno.h>

#include <arm/star/starreg.h>
#include <arm/star/starvar.h>

struct starwdog_softc {
	device_t sc_dev;
	bus_addr_t sc_addr;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	void *sc_ih;
};

static int starwdog_match(device_t, struct cfdata *, void *);
static void starwdog_attach(device_t, device_t, void *);
static void starwdog_init(struct starwdog_softc *);
static int starwdog_intr(void *);

CFATTACH_DECL_NEW(starwdog, sizeof(struct starwdog_softc),
    starwdog_match, starwdog_attach, NULL, NULL);


/* ARGSUSED */
static int
starwdog_match(device_t parent, struct cfdata *match, void *aux)
{
	struct star_attach_args *sa;

	sa = aux;
	sa->sa_size = STRx100_WDT_REGSIZE;

	return 1;
}

/* ARGSUSED */
static void
starwdog_attach(device_t parent, device_t self, void *aux)
{
	struct starwdog_softc *sc;
	struct star_attach_args *sa;

	sa = aux;
	sc = device_private(self);
	sc->sc_dev = self;
	sc->sc_iot = sa->sa_iot;
	sc->sc_addr = sa->sa_addr;

	aprint_normal(": WatchDogTimer\n");
	aprint_naive("\n");

	if (bus_space_map(sc->sc_iot, sc->sc_addr, sa->sa_size, 0,
	    &sc->sc_ioh)) {
		aprint_error(": can't map registers\n");
		return;
	}

	starwdog_init(sc);

	sc->sc_ih = star_intr_establish(STAR_IRQ_WDOG, IPL_HIGH,
	    starwdog_intr, sc);
}

static void
starwdog_init(struct starwdog_softc *sc)
{
}

static int
starwdog_intr(void *arg)
{
	struct starwdog_softc *sc;

	sc = (struct starwdog_softc *)arg;

	return 1;
}
